/*
 * CRC encoder (for codec/format testing)
 * Copyright (c) 2002 Fabrice Bellard
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "libavutil/adler32.h"
#include "avformat.h"

typedef struct CRCState {
    uint32_t crcval;
} CRCState;

static int crc_write_header(struct AVFormatContext *s)
{
#ifdef _MSC_VER /* add cast (1) */
    CRCState *crc = (CRCState *)s->priv_data;
#else /* !_MSC_VER */
    CRCState *crc = s->priv_data;
#endif /* _MSC_VER y/n */

    /* init CRC */
    crc->crcval = 1;

    return 0;
}

static int crc_write_packet(struct AVFormatContext *s, AVPacket *pkt)
{
#ifdef _MSC_VER /* add cast (1) */
    CRCState *crc = (CRCState *)s->priv_data;
#else /* !_MSC_VER */
    CRCState *crc = s->priv_data;
#endif /* _MSC_VER y/n */
    crc->crcval = av_adler32_update(crc->crcval, pkt->data, pkt->size);
    return 0;
}

static int crc_write_trailer(struct AVFormatContext *s)
{
#ifdef _MSC_VER /* add cast (1) */
    CRCState *crc = (CRCState *)s->priv_data;
#else /* !_MSC_VER */
    CRCState *crc = s->priv_data;
#endif /* _MSC_VER y/n */
    char buf[64];

    snprintf(buf, sizeof(buf), "CRC=0x%08x\n", crc->crcval);
#ifdef _MSC_VER /* add cast */
    put_buffer(s->pb, (const unsigned char *)buf, strlen(buf));
#else /* !_MSC_VER */
    put_buffer(s->pb, buf, strlen(buf));
#endif /* _MSC_VER y/n */
    put_flush_packet(s->pb);
    return 0;
}

AVOutputFormat crc_muxer = {
    "crc",
    NULL_IF_CONFIG_SMALL("CRC testing format"),
    NULL,
    "",
    sizeof(CRCState),
    CODEC_ID_PCM_S16LE,
    CODEC_ID_RAWVIDEO,
    crc_write_header,
    crc_write_packet,
    crc_write_trailer,
};
