/*
 * NUT (de)muxing via libnut
 * copyright (c) 2006 Oded Shimon <ods15@ods15.dyndns.org>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file libavformat/libnut.c
 * NUT demuxing and muxing via libnut.
 * @author Oded Shimon <ods15@ods15.dyndns.org>
 */

#include "avformat.h"
#include "riff.h"
#include <libnut.h>

#define ID_STRING "nut/multimedia container"
#define ID_LENGTH (strlen(ID_STRING) + 1)

typedef struct {
    nut_context_tt * nut;
    nut_stream_header_tt * s;
} NUTContext;

static const AVCodecTag nut_tags[] = {
    { CODEC_ID_MPEG4,  MKTAG('m', 'p', '4', 'v') },
    { CODEC_ID_MP3,    MKTAG('m', 'p', '3', ' ') },
    { CODEC_ID_VORBIS, MKTAG('v', 'r', 'b', 's') },
    { 0, 0 },
};

#if CONFIG_LIBNUT_MUXER
static int av_write(void * h, size_t len, const uint8_t * buf) {
#ifdef _MSC_VER /* add cast (1) */
    ByteIOContext * bc = (ByteIOContext *)h;
#else /* !_MSC_VER */
    ByteIOContext * bc = h;
#endif /* _MSC_VER y/n */
    put_buffer(bc, buf, len);
    //put_flush_packet(bc);
    return len;
}

static int nut_write_header(AVFormatContext * avf) {
#ifdef _MSC_VER /* add cast (1) */
    NUTContext * priv = (NUTContext *)avf->priv_data;
#else /* !_MSC_VER */
    NUTContext * priv = avf->priv_data;
#endif /* _MSC_VER y/n */
    ByteIOContext * bc = avf->pb;
#ifdef _MSC_VER /* until MSVC supports dot(.) aligned members, etc */
    nut_muxer_opts_tt mopts = {
       { bc, av_write }, /* nut_output_stream_tt output     */
       { av_malloc, av_realloc, av_free }, /* nut_alloc_tt alloc              */
       1, /* int write_index                 */
       0, /* int realtime_stream             */
       32768, /* int max_distance                */
       NULL /* nut_frame_table_input_tt *  fti */
};
#else /* !_MSC_VER */
    nut_muxer_opts_tt mopts = {
        .output = {
            .priv = bc,
            .write = av_write,
        },
        .alloc = { av_malloc, av_realloc, av_free },
        .write_index = 1,
        .realtime_stream = 0,
        .max_distance = 32768,
        .fti = NULL,
    };
#endif /* _MSC_VER y/n */
    nut_stream_header_tt * s;
    int i;

#ifdef _MSC_VER /* add cast */
    priv->s = s = (nut_stream_header_tt *)av_mallocz((avf->nb_streams + 1) * sizeof*s);
#else /* !_MSC_VER */
    priv->s = s = av_mallocz((avf->nb_streams + 1) * sizeof*s);
#endif /* _MSC_VER y/n */

    for (i = 0; i < avf->nb_streams; i++) {
        AVCodecContext * codec = avf->streams[i]->codec;
        int j;
        int fourcc = 0;
        int num, denom, ssize;

        s[i].type = codec->codec_type == CODEC_TYPE_VIDEO ? NUT_VIDEO_CLASS : NUT_AUDIO_CLASS;

        if (codec->codec_tag) fourcc = codec->codec_tag;
        else fourcc = ff_codec_get_tag(nut_tags, codec->codec_id);

        if (!fourcc) {
            if (codec->codec_type == CODEC_TYPE_VIDEO) fourcc = ff_codec_get_tag(ff_codec_bmp_tags, codec->codec_id);
            if (codec->codec_type == CODEC_TYPE_AUDIO) fourcc = ff_codec_get_tag(ff_codec_wav_tags, codec->codec_id);
        }

        s[i].fourcc_len = 4;
#ifdef _MSC_VER /* add cast (NEW 1) */
        s[i].fourcc = (uint8_t *) av_malloc(s[i].fourcc_len);
#else /* !_MSC_VER */
        s[i].fourcc = av_malloc(s[i].fourcc_len);
#endif /* _MSC_VER y/n */
        for (j = 0; j < s[i].fourcc_len; j++) s[i].fourcc[j] = (fourcc >> (j*8)) & 0xFF;

        ff_parse_specific_params(codec, &num, &ssize, &denom);
        av_set_pts_info(avf->streams[i], 60, denom, num);

        s[i].time_base.num = denom;
        s[i].time_base.den = num;

        s[i].fixed_fps = 0;
        s[i].decode_delay = codec->has_b_frames;
        s[i].codec_specific_len = codec->extradata_size;
        s[i].codec_specific = codec->extradata;

        if (codec->codec_type == CODEC_TYPE_VIDEO) {
            s[i].width = codec->width;
            s[i].height = codec->height;
            s[i].sample_width = 0;
            s[i].sample_height = 0;
            s[i].colorspace_type = 0;
        } else {
            s[i].samplerate_num = codec->sample_rate;
            s[i].samplerate_denom = 1;
            s[i].channel_count = codec->channels;
        }
    }

    s[avf->nb_streams].type = -1;
    priv->nut = nut_muxer_init(&mopts, s, NULL);

    return 0;
}

static int nut_write_packet(AVFormatContext * avf, AVPacket * pkt) {
#ifdef _MSC_VER /* add cast (1) */
    NUTContext * priv = (NUTContext *)avf->priv_data;
#else /* !_MSC_VER */
    NUTContext * priv = avf->priv_data;
#endif /* _MSC_VER y/n */
    nut_packet_tt p;

    p.len = pkt->size;
    p.stream = pkt->stream_index;
    p.pts = pkt->pts;
    p.flags = pkt->flags & PKT_FLAG_KEY ? NUT_FLAG_KEY : 0;
    p.next_pts = 0;

    nut_write_frame_reorder(priv->nut, &p, pkt->data);

    return 0;
}

static int nut_write_trailer(AVFormatContext * avf) {
    ByteIOContext * bc = avf->pb;
#ifdef _MSC_VER /* add cast (1) */
    NUTContext * priv = (NUTContext *)avf->priv_data;
#else /* !_MSC_VER */
    NUTContext * priv = avf->priv_data;
#endif /* _MSC_VER y/n */
    int i;

    nut_muxer_uninit_reorder(priv->nut);
    put_flush_packet(bc);

    for(i = 0; priv->s[i].type != -1; i++ ) av_freep(&priv->s[i].fourcc);
    av_freep(&priv->s);

    return 0;
}

#ifdef _MSC_VER /* until MSVC supports dot(.) aligned members, etc */
AVOutputFormat libnut_muxer = {
    "libnut",           /* const char * name                                                                          */
    "nut format",       /* const char * long_name                                                                     */
    "video/x-nut",      /* const char * mime_type                                                                     */
    "nut",              /* const char * extensions                                                                    */
    sizeof(NUTContext), /* int priv_data_size                                                                         */
    CODEC_ID_VORBIS,    /* enum CodecID audio_codec                                                                   */
    CODEC_ID_MPEG4,     /* enum CodecID video_codec                                                                   */
    nut_write_header,   /* int (*write_header)(struct AVFormatContext *)                                              */
    nut_write_packet,   /* int (*write_packet)(struct AVFormatContext *, AVPacket *pkt)                               */
    nut_write_trailer,  /* int (*write_trailer)(struct AVFormatContext *)                                             */
    AVFMT_GLOBALHEADER, /* int flags                                                                                  */
    NULL,               /* int (*set_parameters)(struct AVFormatContext *, AVFormatParameters *)                      */
    NULL,               /* int (*interleave_packet)(struct AVFormatContext *, AVPacket *out, AVPacket *in, int flush) */
    NULL,               /* const struct AVCodecTag * const * codec_tag                                                */
    CODEC_ID_NONE,      /* enum CodecID subtitle_codec                                                                */
    NULL,               /* const AVMetadataConv * metadata_conv                                                       */
    NULL                /* struct AVOutputFormat * next                                                               */
};
#else /* !_MSC_VER */
AVOutputFormat libnut_muxer = {
    "libnut",
    "nut format",
    "video/x-nut",
    "nut",
    sizeof(NUTContext),
    CODEC_ID_VORBIS,
    CODEC_ID_MPEG4,
    nut_write_header,
    nut_write_packet,
    nut_write_trailer,
    .flags = AVFMT_GLOBALHEADER,
};
#endif /* _MSC_VER y/n */
#endif /* CONFIG_LIBNUT_MUXER */

static int nut_probe(AVProbeData *p) {
    if (!memcmp(p->buf, ID_STRING, ID_LENGTH)) return AVPROBE_SCORE_MAX;

    return 0;
}

static size_t av_read(void * h, size_t len, uint8_t * buf) {
#ifdef _MSC_VER /* add cast (1) */
    ByteIOContext * bc = (ByteIOContext *)h;
#else /* !_MSC_VER */
    ByteIOContext * bc = h;
#endif /* _MSC_VER y/n */
    return get_buffer(bc, buf, len);
}

static off_t av_seek(void * h, long long pos, int whence) {
#ifdef _MSC_VER /* add cast (1) */
    ByteIOContext * bc = (ByteIOContext *)h;
#else /* !_MSC_VER */
    ByteIOContext * bc = h;
#endif /* _MSC_VER y/n */
    if (whence == SEEK_END) {
        pos = url_fsize(bc) + pos;
        whence = SEEK_SET;
    }
    return url_fseek(bc, pos, whence);
}

static int nut_read_header(AVFormatContext * avf, AVFormatParameters * ap) {
#ifdef _MSC_VER /* add cast (1) */
    NUTContext * priv = (NUTContext *)avf->priv_data;
#else /* !_MSC_VER */
    NUTContext * priv = avf->priv_data;
#endif /* _MSC_VER y/n */
    ByteIOContext * bc = avf->pb;
#ifdef _MSC_VER /* until MSVC supports dot(.) aligned members, etc */
    nut_demuxer_opts_tt dopts = {
        { bc, av_read, av_seek, NULL, 0 },  /* nut_input_stream_tt input                                */
        { av_malloc, av_realloc, av_free }, /* nut_alloc_tt alloc                                       */
        1,                                  /* int read_index                                           */
        1,                                  /* int cache_syncpoints                                     */
        NULL,                               /* void * info_priv                                         */
        NULL                                /* void (*new_info)(void * priv, nut_info_packet_tt * info) */
    };
#else /* !_MSC_VER */
    nut_demuxer_opts_tt dopts = {
        .input = {
            .priv = bc,
            .seek = av_seek,
            .read = av_read,
            .eof = NULL,
            .file_pos = 0,
        },
        .alloc = { av_malloc, av_realloc, av_free },
        .read_index = 1,
        .cache_syncpoints = 1,
    };
#endif /* _MSC_VER y/n */
    nut_context_tt * nut = priv->nut = nut_demuxer_init(&dopts);
    nut_stream_header_tt * s;
    int ret, i;

    if ((ret = nut_read_headers(nut, &s, NULL))) {
        av_log(avf, AV_LOG_ERROR, " NUT error: %s\n", nut_error(ret));
        nut_demuxer_uninit(nut);
        return -1;
    }

    priv->s = s;

    for (i = 0; s[i].type != -1 && i < 2; i++) {
        AVStream * st = av_new_stream(avf, i);
        int j;

        for (j = 0; j < s[i].fourcc_len && j < 8; j++) st->codec->codec_tag |= s[i].fourcc[j]<<(j*8);

        st->codec->has_b_frames = s[i].decode_delay;

        st->codec->extradata_size = s[i].codec_specific_len;
        if (st->codec->extradata_size) {
#ifdef _MSC_VER /* add cast (NEW 1) */
            st->codec->extradata = (uint8_t *) av_mallocz(st->codec->extradata_size);
#else /* !_MSC_VER */
            st->codec->extradata = av_mallocz(st->codec->extradata_size);
#endif /* _MSC_VER y/n */
            memcpy(st->codec->extradata, s[i].codec_specific, st->codec->extradata_size);
        }

        av_set_pts_info(avf->streams[i], 60, s[i].time_base.num, s[i].time_base.den);
        st->start_time = 0;
        st->duration = s[i].max_pts;

        st->codec->codec_id = ff_codec_get_id(nut_tags, st->codec->codec_tag);

        switch(s[i].type) {
        case NUT_AUDIO_CLASS:
            st->codec->codec_type = CODEC_TYPE_AUDIO;
            if (st->codec->codec_id == CODEC_ID_NONE) st->codec->codec_id = ff_codec_get_id(ff_codec_wav_tags, st->codec->codec_tag);

            st->codec->channels = s[i].channel_count;
            st->codec->sample_rate = s[i].samplerate_num / s[i].samplerate_denom;
            break;
        case NUT_VIDEO_CLASS:
            st->codec->codec_type = CODEC_TYPE_VIDEO;
            if (st->codec->codec_id == CODEC_ID_NONE) st->codec->codec_id = ff_codec_get_id(ff_codec_bmp_tags, st->codec->codec_tag);

            st->codec->width = s[i].width;
            st->codec->height = s[i].height;
            st->sample_aspect_ratio.num = s[i].sample_width;
            st->sample_aspect_ratio.den = s[i].sample_height;
            break;
        }
        if (st->codec->codec_id == CODEC_ID_NONE) av_log(avf, AV_LOG_ERROR, "Unknown codec?!\n");
    }

    return 0;
}

static int nut_read_packet(AVFormatContext * avf, AVPacket * pkt) {
#ifdef _MSC_VER /* add cast (1) */
    NUTContext * priv = (NUTContext *)avf->priv_data;
#else /* !_MSC_VER */
    NUTContext * priv = avf->priv_data;
#endif /* _MSC_VER y/n */
    nut_packet_tt pd;
    int ret;

    ret = nut_read_next_packet(priv->nut, &pd);

    if (ret || av_new_packet(pkt, pd.len) < 0) {
        if (ret != NUT_ERR_EOF)
            av_log(avf, AV_LOG_ERROR, " NUT error: %s\n", nut_error(ret));
        return -1;
    }

    if (pd.flags & NUT_FLAG_KEY) pkt->flags |= PKT_FLAG_KEY;
    pkt->pts = pd.pts;
    pkt->stream_index = pd.stream;
    pkt->pos = url_ftell(avf->pb);

    ret = nut_read_frame(priv->nut, &pd.len, pkt->data);

    return ret;
}

static int nut_read_seek(AVFormatContext * avf, int stream_index, int64_t target_ts, int flags) {
#ifdef _MSC_VER /* add cast (1) */
    NUTContext * priv = (NUTContext *)avf->priv_data;
#else /* !_MSC_VER */
    NUTContext * priv = avf->priv_data;
#endif /* _MSC_VER y/n */
    int active_streams[] = { stream_index, -1 };
    double time_pos = target_ts * priv->s[stream_index].time_base.num / (double)priv->s[stream_index].time_base.den;

    if (nut_seek(priv->nut, time_pos, 2*!(flags & AVSEEK_FLAG_BACKWARD), active_streams)) return -1;

    return 0;
}

static int nut_read_close(AVFormatContext *s) {
#ifdef _MSC_VER /* add cast (1) */
    NUTContext * priv = (NUTContext *)s->priv_data;
#else /* !_MSC_VER */
    NUTContext * priv = s->priv_data;
#endif /* _MSC_VER y/n */

    nut_demuxer_uninit(priv->nut);

    return 0;
}

#ifdef _MSC_VER /* until MSVC supports dot(.) aligned members, etc */
AVInputFormat libnut_demuxer = {
    "libnut",                           /* const char * name                                                                                                     */
    NULL_IF_CONFIG_SMALL("NUT format"), /* const char * long_name                                                                                                */
    sizeof(NUTContext),                 /* int priv_data_size                                                                                                    */
    nut_probe,                          /* int (*read_probe)(AVProbeData *)                                                                                      */
    nut_read_header,                    /* int (*read_header)(struct AVFormatContext *, AVFormatParameters *ap)                                                  */
    nut_read_packet,                    /* int (*read_packet)(struct AVFormatContext *, AVPacket *pkt)                                                           */
    nut_read_close,                     /* int (*read_close)(struct AVFormatContext *)                                                                           */
    nut_read_seek,                      /* int (*read_seek)(struct AVFormatContext *, int stream_index, int64_t timestamp, int flags)                            */
    NULL,                               /* int64_t (*read_timestamp)(struct AVFormatContext *s, int stream_index, int64_t *pos, int64_t pos_limit)               */
    0,                                  /* int flags                                                                                                             */
    "nut",                              /* const char * extensions                                                                                               */
    0,                                  /* int value                                                                                                             */
    NULL,                               /* int (*read_play)(struct AVFormatContext *)                                                                            */
    NULL,                               /* int (*read_pause)(struct AVFormatContext *)                                                                           */
    NULL,                               /* const struct AVCodecTag * const * codec_tag                                                                           */
    NULL,                               /* int (*read_seek2)(struct AVFormatContext *s, int stream_index, int64_t min_ts, int64_t ts, int64_t max_ts, int flags) */
    NULL,                               /* const AVMetadataConv * metadata_conv                                                                                  */
    NULL                                /* struct AVInputFormat * next                                                                                           */
};
#else /* !_MSC_VER */
AVInputFormat libnut_demuxer = {
    "libnut",
    NULL_IF_CONFIG_SMALL("NUT format"),
    sizeof(NUTContext),
    nut_probe,
    nut_read_header,
    nut_read_packet,
    nut_read_close,
    nut_read_seek,
    .extensions = "nut",
};
#endif /* _MSC_VER y/n */

