/*
 * Creative Voice File muxer.
 * Copyright (c) 2006  Aurelien Jacobs <aurel@gnuage.org>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "voc.h"


typedef struct voc_enc_context {
    int param_written;
} VocEncContext;

static int voc_write_header(AVFormatContext *s)
{
    ByteIOContext *pb = s->pb;
    const int header_size = 26;
    const int version = 0x0114;

    if (s->nb_streams != 1
        || s->streams[0]->codec->codec_type != CODEC_TYPE_AUDIO)
        return AVERROR_PATCHWELCOME;

    put_buffer(pb, ff_voc_magic, sizeof(ff_voc_magic) - 1);
    put_le16(pb, header_size);
    put_le16(pb, version);
    put_le16(pb, ~version + 0x1234);

    return 0;
}

static int voc_write_packet(AVFormatContext *s, AVPacket *pkt)
{
#ifdef _MSC_VER /* add cast (1) */
    VocEncContext *voc = (VocEncContext *)s->priv_data;
#else /* !_MSC_VER */
    VocEncContext *voc = s->priv_data;
#endif /* _MSC_VER y/n */
    AVCodecContext *enc = s->streams[0]->codec;
    ByteIOContext *pb = s->pb;

    if (!voc->param_written) {
        if (enc->codec_tag > 0xFF) {
            put_byte(pb, VOC_TYPE_NEW_VOICE_DATA);
            put_le24(pb, pkt->size + 12);
            put_le32(pb, enc->sample_rate);
            put_byte(pb, enc->bits_per_coded_sample);
            put_byte(pb, enc->channels);
            put_le16(pb, enc->codec_tag);
            put_le32(pb, 0);
        } else {
            if (s->streams[0]->codec->channels > 1) {
                put_byte(pb, VOC_TYPE_EXTENDED);
                put_le24(pb, 4);
                put_le16(pb, 65536-256000000/(enc->sample_rate*enc->channels));
                put_byte(pb, enc->codec_tag);
                put_byte(pb, enc->channels - 1);
            }
            put_byte(pb, VOC_TYPE_VOICE_DATA);
            put_le24(pb, pkt->size + 2);
            put_byte(pb, 256 - 1000000 / enc->sample_rate);
            put_byte(pb, enc->codec_tag);
        }
        voc->param_written = 1;
    } else {
        put_byte(pb, VOC_TYPE_VOICE_DATA_CONT);
        put_le24(pb, pkt->size);
    }

    put_buffer(pb, pkt->data, pkt->size);
    return 0;
}

static int voc_write_trailer(AVFormatContext *s)
{
    put_byte(s->pb, 0);
    return 0;
}

#ifdef _MSC_VER /* until MSVC supports dot(.) aligned members, etc */
static const AVCodecTag * _s_AVC_voc_muxer[] = {ff_voc_codec_tags, 0};
AVOutputFormat voc_muxer = {
    "voc",                                              /* const char * name                                */
    NULL_IF_CONFIG_SMALL("Creative Voice file format"), /* const char * long_name                           */
    "audio/x-voc",                                      /* const char * mime_type                           */
    "voc",                                              /* const char * extensions                          */
    sizeof(VocEncContext),                              /* int priv_data_size                               */
    CODEC_ID_PCM_U8,                                    /* enum CodecID audio_codec                         */
    CODEC_ID_NONE,                                      /* enum CodecID video_codec                         */
    voc_write_header,                                   /* int (*write_header)(struct AVFormatContext *)    */
    voc_write_packet,                                   /* int (*write_packet)(struct AVFormatContext *, AVPacket *pkt) */
    voc_write_trailer,                                  /* int (*write_trailer)(struct AVFormatContext *)   */
    0,                                                  /* int flags                                        */
    NULL,                                               /* int (*set_parameters)(struct AVFormatContext *, AVFormatParameters *) */
    NULL,                                               /* int (*interleave_packet)(struct AVFormatContext *, AVPacket *out, AVPacket *in, int flush) */
    _s_AVC_voc_muxer,                                   /* const struct AVCodecTag * const * codec_tag      */
    CODEC_ID_NONE,                                      /* enum CodecID subtitle_codec                      */
    NULL,                                               /* const AVMetadataConv * metadata_conv             */
    NULL                                                /* struct AVOutputFormat * next                     */
};
#else /* !_MSC_VER */
AVOutputFormat voc_muxer = {
    "voc",
    NULL_IF_CONFIG_SMALL("Creative Voice file format"),
    "audio/x-voc",
    "voc",
    sizeof(VocEncContext),
    CODEC_ID_PCM_U8,
    CODEC_ID_NONE,
    voc_write_header,
    voc_write_packet,
    voc_write_trailer,
    .codec_tag=(const AVCodecTag* const []){ff_voc_codec_tags, 0},
};
#endif /* _MSC_VER y/n */
